#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------
# Imports
# --------------------------------------------------------------------------------------
from datetime import datetime
from typing import Any

from pydantic import BaseModel
from pydantic import Extra

from ramodels.exceptions import ISOParseError

try:
    import zoneinfo
except ImportError:  # pragma: no cover
    from backports import zoneinfo  # type: ignore


# --------------------------------------------------------------------------------------
# Globals
# --------------------------------------------------------------------------------------

# TODO: Perhaps it's worth reading from e.g. env vars here
DEFAULT_TZ = zoneinfo.ZoneInfo("Europe/Copenhagen")

POS_INF = "infinity"
NEG_INF = "-infinity"

# --------------------------------------------------------------------------------------
# Base models
# --------------------------------------------------------------------------------------


class RABase(BaseModel):
    """Base model for RA data models."""

    # TODO: This is duplicated to each class that cannot be instantiated.
    # We should probably find a better solution.
    def __new__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls is RABase:
            raise TypeError("RABase may not be instantiated")
        return super().__new__(cls)

    class Config:
        frozen = True
        allow_population_by_field_name = True
        extra = Extra.forbid


# --------------------------------------------------------------------------------------
# Base utils
# --------------------------------------------------------------------------------------
def tz_isodate(dt: Any) -> datetime:
    """Attempts to parse an incoming value as a timezone aware datetime.

    Args:
        dt (Any): Value to parse into timezone aware datetime.

    Raises:
        ISOParseError: If the incoming value cannot be parsed by dateutil's isoparser.

    Returns:
        Timezone aware datetime object.

        Note that the default {DEFAULT_TZ} is used.
    """
    try:
        iso_dt = datetime.fromisoformat(str(dt))
    except ValueError:
        raise ISOParseError(dt)

    iso_dt = iso_dt if iso_dt.tzinfo else iso_dt.replace(tzinfo=DEFAULT_TZ)
    return iso_dt



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------
# Imports
# --------------------------------------------------------------------------------------
from typing import Optional
from uuid import UUID

from pydantic import Field

from ._shared import EffectiveTime
from ._shared import FacetRef
from ._shared import KlasseAttributes
from ._shared import KlasseProperties
from ._shared import KlasseRelations
from ._shared import KlasseStates
from ._shared import LoraBase
from ._shared import Published
from ._shared import Responsible


# --------------------------------------------------------------------------------------
# Klasse model
# --------------------------------------------------------------------------------------


class Klasse(LoraBase):
    """
    Attributes:
        attributes:
        states:
        relations:
    """

    attributes: KlasseAttributes = Field(alias="attributter")
    states: KlasseStates = Field(alias="tilstande")
    relations: KlasseRelations = Field(alias="relationer")

    @classmethod
    def from_simplified_fields(
        cls,
        facet_uuid: UUID,  # uuid
        uuid: UUID,
        user_key: str,  # rarely used
        organisation_uuid: UUID,
        title: str,
        scope: Optional[str] = None,
        from_date: str = "-infinity",
        to_date: str = "infinity",
    ) -> "Klasse":
        # Inner fields
        _effective_time = EffectiveTime(from_date=from_date, to_date=to_date)
        _properties = KlasseProperties(
            user_key=user_key,
            title=title,
            scope=scope,
            effective_time=_effective_time,
        )
        _published = Published(effective_time=_effective_time)
        _responsible = Responsible(
            uuid=organisation_uuid,
            effective_time=_effective_time,
        )
        _facet = FacetRef(
            uuid=facet_uuid,
            effective_time=_effective_time,
        )

        # Klasse fields
        attributes = KlasseAttributes(properties=[_properties])
        states = KlasseStates(published_state=[_published])
        relations = KlasseRelations(responsible=[_responsible], facet=[_facet])

        return cls(
            attributes=attributes,
            states=states,
            relations=relations,
            uuid=uuid,
        )



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------
# Imports
# --------------------------------------------------------------------------------------
from typing import List
from typing import Literal
from typing import Optional
from uuid import UUID

from .._shared import ManagerLevel
from .._shared import ManagerType
from .._shared import MOBase
from .._shared import OrgUnitRef
from .._shared import PersonRef
from .._shared import Responsibility
from .._shared import Validity

# --------------------------------------------------------------------------------------
# Manager implementations
# --------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------
# Manager model
# --------------------------------------------------------------------------------------


class Manager(MOBase):
    """
    Attributes:
        type:
        org_unit:
        person:
        responsibility:
        manager_level:
        manager_type:
        validity:
    """

    type: Literal["manager"] = "manager"
    org_unit: OrgUnitRef
    person: PersonRef
    responsibility: List[Responsibility]
    manager_level: ManagerLevel
    manager_type: ManagerType
    validity: Validity

    @classmethod
    def from_simplified_fields(
        cls,
        uuid: UUID,
        org_unit_uuid: UUID,
        person_uuid: UUID,
        responsibility_uuids: List[UUID],
        manager_level_uuid: UUID,
        manager_type_uuid: UUID,
        from_date: str,
        to_date: Optional[str] = None,
    ) -> "Manager":
        person = PersonRef(uuid=person_uuid)
        org_unit = OrgUnitRef(uuid=org_unit_uuid)
        responsibility = [
            Responsibility(uuid=r_uuid) for r_uuid in responsibility_uuids
        ]
        manager_level = ManagerLevel(uuid=manager_level_uuid)
        manager_type = ManagerType(uuid=manager_type_uuid)
        validity = Validity(from_date=from_date, to_date=to_date)

        return cls(
            uuid=uuid,
            org_unit=org_unit,
            person=person,
            responsibility=responsibility,
            manager_level=manager_level,
            manager_type=manager_type,
            validity=validity,
        )



#!/usr/bin/env python3
# --------------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2021 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
# --------------------------------------------------------------------------------------
# --------------------------------------------------------------------------------------
# Imports
# --------------------------------------------------------------------------------------
import warnings
from datetime import date
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple

from pydantic import Field
from pydantic import root_validator
from pydantic import validator

from ._shared import MOBase
from ._shared import OrganisationRef
from .details import Details
from ramodels.base import tz_isodate

# --------------------------------------------------------------------------------------
# Employee model
# --------------------------------------------------------------------------------------

# Type aliases
DictStrAny = Dict[str, Any]


# Helper functions
def deprecation(message: str) -> None:
    warnings.warn(message, DeprecationWarning, stacklevel=2)


def split_name(name: str) -> Tuple[str, str]:
    split = name.split(" ", maxsplit=1)
    if len(split) == 1:
        split.append("")
    givenname, surname = split
    return givenname, surname


def validate_names(
    values: DictStrAny, name_key: str, givenname_key: str, surname_key: str
) -> Dict:
    if name_key in values:
        # Both name and given/surname are given erroneously
        if any([givenname_key in values, surname_key in values]):
            raise ValueError(
                f"{name_key} and {givenname_key}/{surname_key} are mutually exclusive"
            )
        # If only name is given, raise a deprecation warning and generate given/surname
        else:
            deprecation(
                f"{name_key} will be deprecated in a future version. "
                f"Prefer {givenname_key}/{surname_key} where possible"
            )
            values[givenname_key], values[surname_key] = split_name(values[name_key])

    return values


class Employee(MOBase):
    """MO Employee data model.

    Attributes:
        type_: Object type.
        givenname: Given name of the employee
        surname: Surname of the employee
        name: Full name of the employee - will be deprecated, use given name/surname
        cpr_no: CPR number of the employee
        seniority: Seniority of the employee
        user_key: Short, unique key identifying the employee
        org: The organisation with which the employee is associated
        nickname_givenname: Given name part of the employee's nickname
        nickname_surname: Surname part of the employee's nickname
        nickname: Full nickname of the employee - will be deprecated,
            use given name/surname
        details:  A list of details to be created for the employee.
    """

    type_: Literal["employee"] = Field("employee", alias="type")
    givenname: str = Field(None)
    surname: str = Field(None)
    name: Optional[str]
    cpr_no: Optional[str] = Field(regex=r"^\d{9}[1-9]$")
    seniority: Optional[datetime]
    user_key: Optional[str]
    org: Optional[OrganisationRef]
    nickname_givenname: Optional[str]
    nickname_surname: Optional[str]
    nickname: Optional[str]
    details: Optional[List[Details]]

    @root_validator(pre=True)
    def validate_name(cls, values: DictStrAny) -> DictStrAny:
        return validate_names(values, "name", "givenname", "surname")

    @root_validator(pre=True)
    def validate_nickname(cls, values: DictStrAny) -> DictStrAny:
        return validate_names(
            values, "nickname", "nickname_givenname", "nickname_surname"
        )

    @validator("cpr_no")
    def validate_cpr(cls, cpr_no: Optional[str]) -> Optional[str]:
        if cpr_no is None:
            return None

        # The string is validated with regex first, so we know this works
        # We only obtain the most significant digit of the code part, since it's
        # what we need for century calculations,
        # cf. https://da.wikipedia.org/wiki/CPR-nummer
        day, month = (cpr_no[0:2], cpr_no[2:4])
        year, code_msd = map(int, (cpr_no[4:6], cpr_no[6]))

        # TODO: let's do pattern matching in 3.10:
        # https://www.python.org/dev/peps/pep-0622/ <3
        century: int = 0
        if code_msd < 4:
            century = 1900
        elif code_msd in {4, 9}:
            if 0 <= year <= 36:
                century = 2000
            else:
                century = 1900
        elif 5 <= code_msd <= 8:
            if 0 <= year <= 57:
                century = 2000
            else:
                century = 1800

        try:
            date.fromisoformat(f"{century+year}-{month}-{day}")
        except Exception:
            raise ValueError("CPR number is invalid.")
        return cpr_no

    @validator("seniority", pre=True, always=True)
    def parse_seniority(cls, seniority: Optional[Any]) -> Optional[datetime]:
        return tz_isodate(seniority) if seniority is not None else None
